# -*- coding: utf-8 -*

"""lsl.common - Common information for the LSL package.  Including:
 * warns    - depercation and experimental feature warnings, 
 * paths    - paths to the module and its data directories, 
 * dp       - information about the DP system, and 
 * stations - information about LWA stations."""

import warns

import paths
import dp
import stations
